interface single_sig_if#(int width);
   reg[width-1:0] o = 'z;
   wire[width-1:0] i;
endinterface

package smp_pkg;
  import uvm_pkg::*;
  import aop_pkg::*;
  `include "uvm_macros.svh"
  
  typedef enum {using_vifs = 0, using_classes = 1} generated_code_t;// used to choose between two kingdoms and normal virtual interfaces
  typedef enum {Verilog = 0, VHDL = 1} signal_hdl;
  typedef enum {Verilog_reg = 0, Verilog_wire = 1} signal_net;

  // A very limited replace function, that will replace a specific char, with a string.
  function automatic string replace_char(string text, string original, string replacement );
    string new_text;

    for(int i = 0;i < text.len(); i ++)
      if (text.substr(i,i) == original)
        new_text = {new_text, replacement};
      else
        new_text = {new_text, text.substr(i,i)};

    return new_text;
  endfunction  

  class smp_aspect extends component_aspect;
    `uvm_object_utils(smp_aspect)
    
    function new(string name = "");
      super.new(name);
    endfunction

     static bit b = add_me();

     static function bit add_me();
	smp_aspect f = new("smp_aspect");
	add_aspect(f);
     endfunction
     
    string hdl_path;

    virtual function void build_hook();
      smp_aspect parent_smp_aspect;
      string local_hdl_path, parent_hdl_path;

      parent_smp_aspect = smp_aspect'(get_aspect(m_parent.get_parent(), this));

      if (parent_smp_aspect != null) 
        parent_hdl_path = parent_smp_aspect.hdl_path;
      
      void'(uvm_config_db#(string)::get(m_parent, "", "hdl_path", local_hdl_path));

      hdl_path = {parent_hdl_path, ((parent_hdl_path == "") || (local_hdl_path == ""))?"":".", local_hdl_path};

    endfunction
  endclass

  typedef class vif_port;

  class smp_port#(int PORT_WIDTH = 1) extends aop_component;

    typedef logic[PORT_WIDTH-1:0] port_type;
    typedef smp_port#(port_type) this_type;

    `uvm_component_param_utils(smp_port#(port_type))

    signal_hdl sig_hdl = Verilog;
    signal_net sig_net = Verilog_reg;

    smp_aspect smp_aspect_ref;

    static int unsigned port_ctr;

    static generated_code_t generated_code = get_generated_code();

    static function generated_code_t get_generated_code();
      string s;
      if ($value$plusargs("GENERATED_CODE=%s", s)) begin
        if (s == "vifs") 
          begin
            get_generated_code = using_vifs;
            set_type_override_by_type(this_type::get_type(), vif_port#(port_type)::get_type());
          end
        else if (s == "classes")
          get_generated_code = using_classes;
      end
    endfunction

    function new(string name, uvm_component parent = null);
      super.new(name, parent);
      smp_aspect_ref = smp_aspect'(get_aspect_by_name("smp_aspect"));
    endfunction

    // these functions are used to access the value of the signal
    virtual function void set(port_type value);
    endfunction

    virtual function port_type get();
    endfunction

    function string get_hdl_path();
      return smp_aspect_ref.hdl_path;
    endfunction

    // get the port configuration parameters
    function void build_phase(uvm_phase phase);
      super.build_phase(phase);
      void'(uvm_config_db#(signal_hdl)::get(this, "", "sig_hdl", sig_hdl));
      void'(uvm_config_db#(signal_net)::get(this, "", "sig_net", sig_net));

     if (port_ctr == 0)
        uvm_report_info("GENERATED_CODE_TYPE", {"connection code is using ", (generated_code == using_classes)?"classes (two kingdoms)": "virtual interfaces"}, UVM_LOW); 

      uvm_report_info("PORT_CONNECTION", {"port ", get_full_name(), " is connected to HDL path: ", smp_aspect_ref.hdl_path}, UVM_LOW) ;

      port_ctr += 1;
    endfunction
 
    // This part is responsible for writing the stub file
    static integer stub_handle = 0;
    integer module_handle = 0;

    function void print_stub_module();
      string module_code;

      if (stub_handle == 0)
        begin
          stub_handle = $fopen("stub.svh", "w");
          if (stub_handle == 0) $finish;
          module_handle = $fopen("stub.sv", "w");
          if (module_handle == 0) $finish;

          module_code = {
            "module stub();\n",
            "  import uvm_pkg::*;\n",
            "  import smp_pkg::*;\n",
            "  `include \"uvm_macros.svh\"\n",
            "  `include \"stub.svh\"\n",
            "endmodule"
          };

          $fdisplay(module_handle, module_code);
        end
    endfunction

    function void print_stub();
      string unique_class_name;
      string stub_code;
      string port_width;

      port_width.itoa(PORT_WIDTH);

      unique_class_name = get_full_name();
      unique_class_name = replace_char(unique_class_name, ".", "___");

      if (generated_code == using_classes)
        stub_code = {
          ((sig_net == Verilog_wire) || (sig_hdl == VHDL))? 
          {"reg[", port_width, "-1:0] ", unique_class_name, "_sig = 'Z;\n", 
           "assign ", smp_aspect_ref.hdl_path, " = ", unique_class_name, "_sig;\n\n"}:"",
          "class ", unique_class_name, " extends smp_port#(", port_width, ");\n",
          "  `uvm_component_utils(",unique_class_name,")\n\n",
          "  function new(string name, uvm_component parent = null);\n",
          "    super.new(name, parent);\n",
          "  endfunction\n\n",
          "  typedef logic[", port_width, "-1:0] port_type;\n\n",
          "  virtual function void set(port_type value);\n",
          (sig_net == Verilog_reg)?
          {"    ",smp_aspect_ref.hdl_path, "= value;\n"}:
          {"    ",unique_class_name, "_sig = value;\n"},
          "  endfunction\n\n",
          "  virtual function port_type get();\n", 
          "    return ", smp_aspect_ref.hdl_path, ";\n",
          "  endfunction\n\n",
          "endclass\n\n",
          "initial\n",
          "  uvm_pkg::factory.set_inst_override_by_type(smp_port#(logic[", port_width, "-1:0])::get_type(), ", unique_class_name, "::get_type(), \"", get_full_name(), "\");\n"
        };                 
      else
        stub_code = {
          "single_sig_if#(", port_width, ") ", unique_class_name, "();\n\n",
          "assign ", unique_class_name, ".i = ", smp_aspect_ref.hdl_path, ";\n",
          ((sig_net == Verilog_wire) || (sig_hdl == VHDL))?
          {"assign ", smp_aspect_ref.hdl_path, " = ", unique_class_name, ".o;\n\n"}:
          {"always @(", unique_class_name, ".o)\n",
          "  if (", unique_class_name, ".o != 'z)\n",
          "    ", smp_aspect_ref.hdl_path, " = ", unique_class_name, ".o;\n\n"},
          "initial\n",
          "  uvm_config_db#(virtual single_sig_if#(", port_width, "))::set(null, \"", get_full_name(),  "\", \"port\", ", unique_class_name, ");\n"
        };
  
      $fdisplay(stub_handle, stub_code);

    endfunction

    // using the connect/end_of_elaboration phases only because one comes after the other
    function void connect_phase(uvm_phase phase);
      string bla;

      super.connect_phase(phase);
      
      if ($test$plusargs("GENERATE_STUB")) begin
        print_stub_module();
        print_stub();
      end
    endfunction

    function void end_of_elaboration_phase(uvm_phase phase);
      string bla;

      super.end_of_elaboration_phase(phase);

      if ($test$plusargs("GENERATE_STUB")) begin
        port_ctr -= 1;
        if (port_ctr == 0) begin
          if (stub_handle != 0)
            $fclose(stub_handle);

            uvm_report_info("STUB_GENERATION_COMPLETED", "stub generation was completed successfully" , UVM_LOW); 
          $finish();
        end
      end
    endfunction
  endclass

  class vif_port#(int PORT_WIDTH = 1) extends smp_port#(PORT_WIDTH);

    typedef logic[PORT_WIDTH-1:0] port_type;
    typedef vif_port#(port_type) this_type;

    `uvm_component_param_utils(this_type)

    virtual single_sig_if#(PORT_WIDTH) port;

    function new(string name, uvm_component parent = null);
      super.new(name, parent);
    endfunction

    // these functions are used to access the value of the signal
    virtual function void set(port_type value);
      port.o = value;
    endfunction

    virtual function port_type get();
      return port.i;
    endfunction

    function void build_phase(uvm_phase phase);
      super.build_phase(phase);
      void'(uvm_config_db#(virtual single_sig_if#(PORT_WIDTH))::get(this, "", "port", port));
    endfunction
  endclass
endpackage