package aop_pkg;
  import uvm_pkg::*;
  `include "uvm_macros.svh"

  typedef class component_aspect;
  typedef class aop_component;
  typedef class aop_env;
  typedef class aop_test;
      
  function automatic component_aspect get_aspect(uvm_component comp, component_aspect required_aspect);
    component_aspect aspects[];

    uvm_root r;
    aop_component c;
    aop_env e;
    aop_test t;

    // uvm_root is instantiated by UVM so we can't extend it (or maybe we can? is it taken from the factory?)
    if ($cast(r, comp))
      return null;

    // probably we could save this case by calling get_type recursively until we get a match with one of the aop_* types
    case(1)
      $cast(c, comp) : aspects = c.m_aspects;
      $cast(e, comp) : aspects = e.m_aspects;
      $cast(t, comp) : aspects = t.m_aspects;
    endcase

    foreach(aspects[i])
      if (aspects[i].get_type() == required_aspect.get_type())
        return aspects[i];
  endfunction


// add AOP capabilities to components by extending every component with a list of aspects. Aspects can be added/removed by adding them to a static queue
  class component_aspect extends uvm_object;
    `uvm_object_utils(component_aspect)    

    uvm_component m_parent;

    function new(string name = "");
      super.new(name);
    endfunction
     
    function component_aspect replicate(uvm_component parent);
      replicate = component_aspect'(clone());
      replicate.m_parent = parent;
    endfunction
  
    // sync aspect to phases so it can do things in a phase dependent way.
    virtual function void build_hook();
    endfunction // build_hook
   
    virtual function void connect_hook();
    endfunction
   
    virtual function void end_of_elaboration_hook();
    endfunction	
   
  endclass
   
  component_aspect aspects[$];

  function void add_aspect(component_aspect aspect);
    aspects[$+1] = aspect;
  endfunction

`define aspect_support \
    \
    component_aspect m_aspects[];\
    \
    function new(string name, uvm_component parent = null);\
      super.new(name, parent);\
      \
      m_aspects = new[aspects.size()];\
      \
      foreach(aspects[i])\
        m_aspects[i] = aspects[i].replicate(this);\
    endfunction\
    \
    function void build_phase(uvm_phase phase);\
      super.build_phase(phase);\
      foreach(m_aspects[i])\
        m_aspects[i].build_hook();\
    endfunction\
    \
    function void connect_phase(uvm_phase phase);\
      super.connect_phase(phase);\
      foreach(m_aspects[i])\
        m_aspects[i].connect_hook();\
    endfunction\
    \
    function void end_of_elaboration_phase(uvm_phase phase);\
      super.end_of_elaboration_phase(phase);\
      foreach(m_aspects[i])\
        m_aspects[i].end_of_elaboration_hook();\
    endfunction\
    \
    function component_aspect get_aspect_by_name(string name);\
      foreach(aspects[i])\
        if (m_aspects[i].get_name() == name)\
          return m_aspects[i];\
    endfunction
   
/*
  // I didn't extend all uvm_component derived classes, only the ones most used
    class aop_component extends uvm_component;
    component_aspect m_aspects[];
    
    function new(string name, uvm_component parent = null);
      super.new(name, parent);
      
      m_aspects = new[aspects.size()];
      
      foreach(aspects[i])
        m_aspects[i] = aspects[i].replicate(this);
    endfunction
    
    function void build_phase(uvm_phase phase);
      super.build_phase(phase);
      foreach(m_aspects[i])
        m_aspects[i].build_hook();
    endfunction
    
    function void connect_phase(uvm_phase phase);
      super.connect_phase(phase);
      foreach(m_aspects[i])
        m_aspects[i].connect_hook();
    endfunction
    
    function void end_of_elaboration_phase(uvm_phase phase);
      super.end_of_elaboration_phase(phase);
      foreach(m_aspects[i])
        m_aspects[i].end_of_elaboration_hook();
    endfunction\
    
    function component_aspect get_aspect_by_name(string name);
      foreach(aspects[i])
        if (m_aspects[i].get_name() == name)
          return m_aspects[i];
    endfunction
  endclass
*/


  class aop_component extends uvm_component;
    `aspect_support
  endclass

  class aop_env extends uvm_env;
    `aspect_support
  endclass

  class aop_test extends uvm_test;
    `aspect_support
  endclass

  class aop_agent extends uvm_agent;
    `aspect_support
  endclass

  class aop_monitor extends uvm_monitor;
    `aspect_support
  endclass

  class aop_scoreboard extends uvm_scoreboard;
    `aspect_support
  endclass

  class aop_sequencer #(type REQ=uvm_sequence_item, RSP=REQ) extends uvm_sequencer#(REQ, RSP);
   `aspect_support
  endclass
endpackage