// testbench code
package tb;
  import uvm_pkg::*;
  `include "uvm_macros.svh"

  typedef enum {type_A, type_B, type_C, PLACEHOLDER1} layer1_t;
  typedef enum {type_X, type_Y, PLACEHOLDER2} layer2_t;
  
  
  class a_fields extends uvm_object;
    function new(string name);
      super.new(name);
    endfunction

    rand int A1, A2, A3;
  endclass
  
  class b_fields extends uvm_object;
    function new(string name);
      super.new(name);
    endfunction

    rand int B1, B2;
  endclass
  
  class c_fields extends uvm_object;
    function new(string name);
      super.new(name);
    endfunction

    rand int C1, C2, C3, C4;
  endclass
  
  class x_fields extends uvm_object;
    function new(string name);
      super.new(name);
    endfunction

    rand int X1, X2, X3;
  endclass
  
  class y_fields extends uvm_object;
    function new(string name);
      super.new(name);
    endfunction

    rand int Y1, Y2, Y3;
  endclass
  
  
  class packet extends uvm_sequence_item;
    `uvm_object_utils(packet)
  
    rand layer1_t layer1;
    rand layer2_t layer2;
  
  
    // disabled from test
    constraint not_placeholder1_c { layer1 != PLACEHOLDER1; };
    constraint not_placeholder2_c { layer2 != PLACEHOLDER2; };
  
    rand a_fields A;
    rand b_fields B;
    rand c_fields C;
    rand x_fields X;
    rand y_fields Y;
  
    function void pre_randomize();
      A = new("A");
      B = new("B");
      C = new("C");
      X = new("X");
      Y = new("Y");
    endfunction
  
    function void post_randomize();
      if (layer1 != type_A) A = null;
      if (layer1 != type_B) B = null;
      if (layer1 != type_C) C = null;
      if (layer2 != type_X) X = null;
      if (layer2 != type_Y) Y = null;
    endfunction
   
    constraint B_implies_X_c {layer1 == type_B -> layer2 == type_X; };
    constraint X1_lt_10_when_A_c { ((layer1 == type_A) && (layer2 == type_X)) -> X.X1 < 10; };
    constraint X1_lt_5_when_B_c { ((layer1 == type_B) && (layer2 == type_X)) -> X.X1 < 5; };
    constraint X1_lt_C2_when_C_c { ((layer1 == type_C) && (layer2 == type_X)) -> X.X1 < C.C2; };
  
  endclass : packet
    
  class env extends uvm_env;
    `uvm_component_utils(env)
  
    function new(string name, uvm_component parent);
      super.new(name, parent);
    endfunction
  
    task run_phase(uvm_phase phase);
      packet p;
      int i;
     
      phase.raise_objection(this, "env");
  
      p = packet::type_id::create("packet");
  
      for (i=0; i<10; i++)
        begin
          p.randomize();
          $display("packet #%d is %p", i, p);
        end
  
      phase.drop_objection(this, "env");
    endtask
  endclass
  
  class base_test extends uvm_test;
    `uvm_component_utils(base_test)
  
    env env1;
  
    function new(string name, uvm_component parent);
      super.new(name, parent);
    endfunction
  
    function void build_phase(uvm_phase phase);
      super.build_phase(phase);
  
      env1 = env::type_id::create("env1", this);
    endfunction
  endclass
endpackage

package tests;
  import uvm_pkg::*;
  import tb::*;
  `include "uvm_macros.svh"

  typedef enum {type_D} extra_packet_types_t;

  class d_fields extends uvm_object;
    int D1;
  endclass

  
  // the extended packet can generate all previous subtypes plus the new one
  
  class extended_packet extends packet;
    `uvm_object_utils(extended_packet)
  
    rand extra_packet_types_t extra_type;
  
    rand d_fields D;

    function void pre_randomize();
      super.pre_randomize();
      D = new();
    endfunction

    function void post_randomize();
      super.post_randomize();
   
      if (layer1 != PLACEHOLDER1) D = null;
    endfunction
  
    // disable constraint
    constraint not_placeholder1_c {};
    constraint placeholder_is_d {(layer1 == PLACEHOLDER1) -> (extra_type == type_D);};
  
    constraint some_constraint {((extra_type == type_D) && (layer2 == type_X)) -> X.X1 < D.D1; };
  endclass : extended_packet


  class test_A extends base_test;
    `uvm_component_utils(test_A)
  
    function new(string name, uvm_component parent);
      super.new(name, parent);
    endfunction
  
    function void build_phase(uvm_phase phase);
      super.build_phase(phase);
  
      set_type_override_by_type( packet::get_type(), extended_packet::get_type(), "*");
    endfunction
  endclass
endpackage

module top();
  import uvm_pkg::*;
  import tb::*;
  import tests::*;

  initial
    run_test();
endmodule





  


